module EIO(readRawFile,writeRawFile, makeBitHandle, getBits, putBits, BitHandle, endBitHandle, putRaw, readRaw,atomicWriteFile, getUniqueName, atomicWrite, getTempFileName, memoIO, withTempfile) where

import Bits
import Char
import Control.Concurrent.MVar
import Control.Exception as E
import Data.Array.IO
import Data.Unique
import Directory(removeFile)
import IO
import IOExts
import Monad(liftM)
import Posix
import Word
import System.IO


bufSize = 4096

readRawFile :: String -> IO [Word8]
readRawFile fn = E.bracket (openBinaryFile fn ReadMode) hClose hGetRawContents 

writeRawFile :: String -> [Word8] -> IO ()
writeRawFile fn xs = E.bracket (openBinaryFile fn WriteMode) hClose $ \h -> hPutRawContents h xs

hGetRawContents :: Handle -> IO [Word8]
hGetRawContents h = do
    a <- newArray_ (0,bufSize)
    getall a where
	getall a = do
	    sz <- hGetArray h a bufSize
	    av <- getElems a
	    if sz == 0 then return [] else do
		r <- getall a
		return (take sz av ++ r)
    
hPutRawContents :: Handle -> [Word8] -> IO ()
hPutRawContents h xs = do
    a <- newArray_ (0,bufSize)
    prc a h xs where
	prc _ _ [] = return ()
	prc a h xs@(_:_) = do
	    let (ys,zs) = splitAt bufSize xs
	    let lys = length ys
	    mapM_ (\(i,e) -> writeArray a i e) (zip [0..lys - 1] ys)
	    hPutArray h a lys
	    prc a h zs

data StrictPair = SP !Int !Int

data BitHandle = BitHandle !Handle !(MVar StrictPair) !(MVar Int) (IOUArray Int Word8)



makeBitHandle :: Handle -> IO BitHandle
makeBitHandle h = do
    r <- newMVar (SP 0 0) 
    a <- newArray_ (0,bufSize)
    rv <- newMVar 0
    return $ BitHandle h r rv a

endBitHandle :: BitHandle -> IO ()
endBitHandle (BitHandle h r _ _) = do
    (SP w8 c) <- takeMVar r 
    case c of 
	0 -> return ()
	n -> hPutChar h (chr w8)

putBits :: BitHandle -> Int -> Word32 -> IO ()
putBits _ n _ | n < 0 || n > 32 = error $ "putBits on invalid count: " ++ show n
putBits _ 0 _ = return ()
putBits bh@(BitHandle h r _ _) nbits val = do
    putBit (fromIntegral val `mod` 2)
    putBits bh (nbits - 1) (val `shiftR` 1) where
	putBit b = do
	    (SP w8 c) <- takeMVar r
	    let w8' = (.|.) w8 (shiftL  b  c)
	    case c of 
		7 -> hPutChar h (chr $! w8') >> putMVar r (SP 0 0)
		_ -> putMVar r (SP w8' (c + 1))

getBits :: BitHandle -> Int -> IO Word32
getBits _ nbits | nbits < 0 || nbits > 32 = error $ "getBits on invalid count: " ++ show nbits
getBits _ 0 = return 0
getBits bh@(BitHandle h r rv a) nbits = do
    f nbits where
	f 0 = return 0
	f n = do 
	    b <- getBit
	    r <- f $! (n - 1)
	    return $! (r `shiftL` 1) + fromIntegral b
	getBit = do
	    (SP w8 c) <- takeMVar r
	    if c > 0 then 
		(putMVar r $! (SP (w8 `shiftR` 1)  (c - 1))) >> (return $! (w8 .&. 1))
	     else do
		v <- takeMVar rv
		w8 <- case v of
		    0 -> do
			hGetArray h a bufSize
			putMVar rv (bufSize - 1)
			readArray a 0
		    n -> do
			putMVar rv (n - 1) 
			readArray a (bufSize - n)
		putMVar r $! (SP (fromIntegral w8) 8)
		getBit


putRaw :: Handle -> [Word8] -> IO ()
putRaw h v = hPutStr h (map (chr . fromIntegral) v)

readRaw :: Handle -> Int -> IO [Word8]
readRaw h 0 = return []
readRaw h n = do 
    a <- newArray_ (0,(n - 1))
    sz <- hGetArray h a n
    av <- (getElems a)
    return $! (take sz av)
-- v <- replicateM n (hGetChar h) 
--    return $ map (fromIntegral . ord) v


atomicWrite :: String -> (Handle -> IO a) -> IO a
atomicWrite fn action = do
    n <- getUniqueName
    let tn = fn ++ "." ++ n
    v <- E.bracket (openBinaryFile tn WriteMode) hClose action 
    rename tn fn
    return v
    

atomicWriteFile :: String -> String -> IO ()
atomicWriteFile name s = do
    n <- getUniqueName
    let tn = name ++ "." ++ n
    writeFile tn s
    rename tn name


getUniqueName :: IO String
getUniqueName = do
    id <- getProcessID
    u <- newUnique
    n <- liftM nodeName getSystemID 
    t <- epochTime
    return $ n ++ "." ++ show id ++ "." ++ show t ++ "." ++ show (hashUnique u)
    

memoIO :: IO a -> IO a
memoIO ioa = do
    v <- readIORef var
    case v of 
	Just x -> return x
	Nothing -> do
	    x <- ioa
	    writeIORef var (Just x) 
	    return x 
     where
        {-# NOTINLINE var #-}
	var = unsafePerformIO $ newIORef Nothing

getTempFileName :: IO String
getTempFileName = do
    u <- getUniqueName
    return $ "/tmp/" ++ u


withTempfile :: (String -> IO a) -> IO a
withTempfile action = E.bracket getTempFileName removeFile action

