module Main(main) where

import Array hiding(bounds)
import Char
import IO
import List
import Maybe
import Monad
import System
import System.Time
import Time
import Directory

import Control.Concurrent
import Control.Exception
import Data.IORef
import Data.Unique
import qualified Posix

import Charset
import GinsuConfig
import ConfigFile 
import Curses
import EIO
import ErrorLog
import Filter
import Gale
import GenUtil
import Help
import KeyName
import Puff
import Screen
import Version
import Locale
import SHA1
import CacheIO
import Options
import PackedString
import qualified Data.HashTable as Hash







idleThreshold = 240
pufflog = "ginsu.3.pufflog"
helpText = "Need Help? Press F1 or ?"

getTmpFile = do
    pid <- Posix.getProcessID
    u <- newUnique
    return ("/tmp/ginsu.puff." ++ show pid ++ "." ++ show (hashUnique u))

insertKeys ic s = mapM_ (\c -> writeChan ic (Left c)) (stringToKeys s)

    

    
{-# NOTINLINE main #-}
main = do
    setLocaleAll ""
    configureGinsu
    (flags, acats) <- ginsuOpts
    checkConfigIsGood

    withErrorMessage bugMsg $ do

    logname <- case envErrorLog flags of 
	Just x -> return x
	Nothing -> galeFile "ginsu.errorlog" 

    withErrorLog logname $ withStartEndEntrys "Ginsu" $ do

    case (envVerbose flags) of
	1 -> do 
	    setLogLevel LogInfo
	    putLog LogNotice $ "Verbosity Level: Info" 
	n | n > 1 -> do
	    setLogLevel LogDebug
	    putLog LogNotice $ "Verbosity Level: Debug" 
	_ -> return ()

    cs <- configLookup "CHARSET"
    charsetSetup cs

    setErrorLogPutStr Charset.csHPutStr

    putLog LogNotice $ "CHARSET: " ++ show cs
    

    galeDir <- getGaleDir
    gp <- getGaleProxy
    gid <- getGaleId
    galeDomain <- getGaleDomain
    galeAliases <- getGaleAliases

    putLog LogInfo $ "GALE_DIR " ++ galeDir
    putLog LogInfo $ "GALE_DOMAIN " ++ galeDomain
    putLog LogInfo $ "GALE_ID " ++ gid
    putLog LogInfo $ "GALE_PROXY " ++ simpleQuote gp
    putLog LogInfo $ "GALE_ALIASES \n" ++ 
	unlines ( map  (\(l,(c,d)) -> "  " ++ l ++ " -> " ++ c ++ "@" ++ d)  galeAliases) 


    Control.Exception.bracket_ initCurses endWin $ do
    withCursor CursorInvisible $ do
    -- withMouseEventMask [ButtonClicked 1] $ do

    ic <- newChan
    b <- configLookupBool "DISABLE_SIGWINCH"
    
    redraw_r <- newRenderContext

    Posix.installHandler Posix.sigINT Posix.Ignore Nothing
    Posix.installHandler Posix.sigPIPE Posix.Ignore Nothing
    case (b,cursesSigWinch) of 
	(False,Just s) -> Posix.installHandler s (Posix.Catch (resizeRenderContext redraw_r (writeChan ic (Left KeyResize)))) Nothing >> return ()
	_ -> return ()

    gs <- fmap (concat . map words) $ configLookupList "GALE_SUBSCRIBE"
    gps <- getGaleProxy
    let c =  if envJustArgs flags then acats else gs ++ acats                                            
        nc = map ("_gale.notice@" ++) $ snub (snds (map catParseNew c))  
    --let c =  gs ++ acats
    --    nc = map ("_gale.notice@" ++) $ snub (snds (map catParseNew c))
    withGale gps $ \gc -> do
    galeAddCategories gc $ map catParseNew (snub $ c ++ nc)
    (name,domain) <- fmap catParseNew getGaleId
    galeAddCategories gc $ [("_gale.rr." ++ name ,domain)]
    doRender widgetEmpty
    pl <- forkIO $ puffLoop ic gc
    gl <- forkIO $ getchLoop ic
    yo <- newIORef (0::Int)
    plog <- galeFile pufflog
    ops <- if envNoPufflog flags then return [] else 
        doRender (widgetCenter $  widgetText "Reading pufflog...") >> readPuffs plog
    ps <- newMVar ((reverse $ zip [1..] (reverse ops)))
    mark_set_r <- newIORef (listArray (0,9) (replicate 10 []))
    mapM_ (\n -> checkMark n >>= \f -> mapVal mark_set_r f) [1..9]
    next_r <- newIORef (length ops + 1)
    unless (envNoWritePufflog flags || envNoPufflog flags) $ (forkIO $ pufflogLoop ps 0) >> return ()

    (Just s) <- configLookup "ON_STARTUP" 
    insertKeys ic s 
    
    doRender (widgetCenter $  widgetText "Entering mainloop...") 
    mainLoop gc ic yo ps next_r mark_set_r redraw_r 
    killThread pl
    killThread gl
    ps <- liftM (map snd) $ readVal ps
    unless (envNoPufflog flags || envNoWritePufflog flags) $ doRender (widgetCenter $ widgetText "Writing pufflog...") >>  writePufflog ps

pufflogLoop ps_r n = do
    threadDelay 30000000
    mv <- fmap (fmap fst . listToMaybe) $ readVal ps_r
    case mv of
	Just n' | n' /= n -> do
	    ps <- liftM (map snd) $ readVal ps_r
	    writePufflog ps
	    pufflogLoop ps_r n'
	_ -> pufflogLoop ps_r n

writePufflog ps = do
    plog <- galeFile pufflog
    plsize <- fmap (read . fromJust) $ configLookup "PUFFLOG_SIZE" 
    withPrivateFiles $ writePuffs plog (take plsize ps)


checkMark n = do
    v <- configLookup ("MARK_" ++ show n)
    case v of 
	Nothing -> return id 
	Just z -> case parseFilter z of
	    Nothing -> putLog LogError ("Could not parse mark:\n  " ++ z) >> return id
	    Just y -> return (// [(n - 1,[y])])


puffLoop ic gc = repeatM_ $  galeNextPuff gc >>= \p -> (writeChan ic $ Right p)
getchLoop ic = repeatM_ (getCh >>= \x -> writeChan ic (Left x)) 


apHead f (x:xs) = f x:xs
apHead _ [] = []


expandAliases :: [Category] -> IO [Category]
expandAliases cats = do
    as <- getGaleAliases 
    gd <- getGaleDomain
    let ec x@(c,"") = de $ head $ [(tc ++ drop (length f) c,td) | (f,(tc,td)) <- as, f `isPrefixOf` c] ++ [x]
	ec x = x
	de (c,"") = (c,gd)
	de x = x
    return $ map ec cats 

{-
statusBarWidget :: SVar String -> SVar Int -> SVar Int -> Widget
statusBarWidget svm svs svpc = widgetHorizontalBox False [(NoExpand,m), (Expand, widgetEmpty), (NoExpand,s), (NoExpand,widgetText "/"), (NoExpand, pc), (NoExpand,widgetText " ") ] where
    m = newSVarWidget svm widgetText
    s = newSVarWidget svs (widgetText . show) 
    pc = newSVarWidget svpc (widgetText . show)
-}

type PresenceData = [(String,(String,String))]
presenceString (_,(_,n)) = n

--presenceView :: PresenceData -> Widget
presenceView idleHash pl = dynamicWidget $ do
    itl <- Hash.toList idleHash
    (TOD ct _) <- getClockTime
    let pt = unlines $ "Online:": buildTableLL (concatMap rt (gf  plon)) ++ ["", "Offline:"] ++ buildTableLL (concatMap rt (gf ploff))
        gf ts = map (\ts -> (fst (head ts), snds ts)) (groupBy (\(a,_) (b,_) -> a == b) (sort ts))
        (ploff,plon) = partition (\p -> "out" `isPrefixOf` presenceString p ) pl
        rt (a,bs) = [(a,it a ++ concatInter ", " (snub (snds bs)))]
        it a = case lookup a itl of
            Just n | n == epoch -> "[not idle]"
            Just (TOD nct _) ->  if (ct - nct) < idleThreshold then "[not idle] " else "[" ++ showDuration (ct - nct) ++ "] "
            Nothing -> ""
    return $ widgetText pt 



presenceViewUser :: String -> PresenceData -> Widget
presenceViewUser user pl = widgetText pt where
    pt = user ++ ":\n" ++ indentLines 2  (unlines $ buildTableLL ([b|(a,b) <- pl, a == user]))

	
    
dialog s = widgetCenter $ widgetAttr [AttrBold] $ widgetSimpleFrame wb where
    wb = widgetText $ unlines [les, es, les]
    les = replicate (length es) ' '
    es = (" " ++ s ++ " " )

generateIdInstance = do
    pid <- Posix.getProcessID
    n <- liftM Posix.nodeName Posix.getSystemID
    return $ n ++ "/" ++ show pid

puffTemplate :: IO Puff
puffTemplate = do
    idText <- generateIdInstance
    gi <- getGaleId
    t <- getClockTime
    u <- newUnique
    gn <- configLookup "GALE_NAME"
    n <- case gn of
	Just n -> return ((f_messageSender,FragmentText (packString n)):)
	Nothing -> return id
    let mid = sha1String $ show t ++ " " ++ idText ++ " " ++ show (hashUnique u)
    let sig = [Signed (Key gi [])]
    let ef = n [
	    (f_messageId, FragmentText $ packString mid),
	    (packString "id/class",FragmentText $ packString (package ++ "/" ++ version) ),
	    (packString "id/instance", FragmentText $ packString idText),
	    (f_idTime,FragmentTime t)] 
    return emptyPuff {fragments = ef, signature = sig}

mainLoop gc ic yor psr next_r mark_set_r rc = do 
    hw <- helpWidget 

    message_r <- newMVar helpText
    selected_r <- newMVar 0
    let puffcount_r = readVal psr >>= return . length
    filter_r <- newMVar []

    presence_r <- newMVar []

    presence <- configLookupElse "GALE_PRESENCE" "in.perhaps"
    myPresence <- newJVar presence

    idleHash <- Hash.new (==) Hash.hashString
    puffRRHash <- Hash.new (==) Hash.hashString
    
    presenceWidget <- widgetScroll (newSVarWidget presence_r (presenceView idleHash))

    ct <- getClockTime 
    userActionTime <- newJVar ct
    --resizeSignal <- newSignal

    keyTable <- buildKeyTable

    let fsw = newSVarWidget filter_r $ \fs -> 
	    widgetHorizontalBox False (intersperse (NoExpand, widgetText " ") $ map (\w -> (NoExpand,widgetAttr [AttrReverse] (widgetText w))) (reverse $ map showFilter fs)) 

    selPuff <- newCacheIO $ do
	    ps <-  readVal psr
	    selected <- cacheIO $ readVal selected_r
	    return $ lookup selected ps 


    getRendered <- newCacheIO $ do
        (_,xs) <- cacheIOeq scrSize
        ps <- cacheIO $ readVal psr 
        let nps =  [(x,rPuff xs p)| x@(_,p) <- ps] 
        return nps
    getFiltered <- newCacheIO $ do
        nps <- cacheIO getRendered
        fs <- cacheIO $ readVal filter_r
        let f p = all (`filterAp` p) fs
        return [x| x@((_,p),_) <- nps, f p]

    buf_size <- newCacheIO $ cacheIO getFiltered >>= \ps -> return $ sum [x| (_,(x,_)) <- ps]

    let sb = dynamicWidget $ do
        m <- fmap widgetText $ readVal message_r
        s <- fmap (widgetText . show) $ readVal selected_r
        pc <- fmap (widgetText . show) $  puffcount_r
        st <- readVal $ connectionStatus gc 
        let f = case st of 
                Right _ -> id
                Left s -> \w -> widgetVerticalBox False [(NoExpand,w), (NoExpand,widgetAttr [AttrReverse] (widgetText $ "*** " ++ s ++ " ***"))]
        return $ f $ widgetHorizontalBox False [(NoExpand,m), (Expand, widgetEmpty), (NoExpand,s), (NoExpand,widgetText "/"), (NoExpand, pc), (NoExpand,widgetText " ") ] 
    let	keyError s k = setMessage (s ++ ": " ++ keysToString [k] ++ " -- " ++ helpText)
	setMessage m = writeVal message_r m 
	scrollPuffs x = mapVal yor (+ x) 
	gfs n = do
	    mark_set <- readVal mark_set_r
	    modifyFilter (const (mark_set ! n))
	filter_thread = do
	    p <-  selPuff
	    case p of 
		Just (Puff {cats = [c]}) -> modifyFilter (FilterCategory c:)
		Just (Puff {cats = cs}) -> modifyFilter (FilterOr (map FilterCategory cs):)
		_ -> return ()
	    
	addPuff p = do
            let author = getAuthor p
            (name,domain) <- fmap catParseNew getGaleId
	    case getFragmentString p (packString "notice/presence") of 
		Nothing -> return ()
		Just pn -> do
		    let np@(a',(b',_)) = (author, (maybe "unknown" unpackPS (getFragmentString p (packString "id/instance")), unpackPS pn))
		    mapVal presence_r (\xs -> (np:[x| x@(a,(b,_)) <- xs, a /= a' || b /= b']))
	    case getFragmentTime p (packString "status.idle") of 
		Nothing -> return ()
                Just pn | pn == epoch -> Hash.delete idleHash author >> getClockTime >>=  Hash.insert idleHash author 
		Just pn -> do
                    mt <- Hash.lookup idleHash author
                    Hash.delete idleHash author
                    Hash.insert idleHash author $ fromJust $ max mt (Just pn)
            case getFragmentString p (packString "answer.receipt") of 
                Just n | [(cat,dom)] <- cats p, dom == domain, ("_gale.rr." ++ name) `isPrefixOf` cat -> do
                    mv <- Hash.lookup puffRRHash (drop (10 + length name) cat) 
                    Hash.delete puffRRHash (drop (10 + length name) cat)
                    Hash.insert puffRRHash (drop (10 + length name) cat) (unpackPS n:concat (maybeToMonad mv))
                _ -> return ()
                    
                
	    case getFragmentString p (packString "question.receipt") of
		Nothing -> return ()
		Just _ -> buildReciept p >>= \p -> forkIO (galeSendPuff gc p) >> return ()
	    case fmap unpackPS $ getFragmentString p f_messageBody of
		Nothing -> return ()
		Just n -> when (not $ (all isSpace n) && (all (("_gale" `isPrefixOf`) . fst) (cats p))) $ do
                    mt <- Hash.lookup idleHash author
                    pn <- getClockTime
                    Hash.insert idleHash author $ fromJust $ max mt (Just pn)
		    v <- configLookupList "BEEP"
		    let f = FilterOr (catMaybes (map parseFilter v))
		    when (filterAp f p) Curses.beep
		    n <- readVal next_r
		    mapVal next_r (+1)
		    mapVal psr (\ps -> ((n,p):ps)) 
		    select_perhaps select_next
		    touchRenderContext rc
	composePuff :: IO () -> [Category] -> [PackedString] -> IO ()
	composePuff done cs kwds = do
	    p <- puffTemplate 
	    ep <- editPuff (p {cats = cs, fragments = fragments p ++ [ (f_messageKeyword,FragmentText k) | k <- kwds]})
	    case ep of 
		Nothing -> setMessage "Puff aborted."
		Just p -> do
		    pcw <- puffConfirm addPuff gc done p
		    setRenderWidget rc pcw
	getPresenceFrags = do
	    mp <- readVal myPresence
	    (TOD ct _) <- getClockTime
	    uaT@(TOD uat _) <- readVal userActionTime
	    (it,is,notIdle) <- return $ case ct - uat of 
		n | n < idleThreshold -> (epoch," (not idle)",True)
		n -> (uaT," (" ++ (showDuration n) ++ " idle)",False)
	    return ([
		    (packString "notice/presence",FragmentText (packString (mp ++ is))), 
		    (packString "status.presence", FragmentText $ packString mp), 
		    (packString "status.idle", FragmentTime it)],notIdle)
	buildReciept p = do
	    np <- puffTemplate
	    r <- getFragmentString p (packString "question.receipt")
	    gid <- getGaleId
	    let rid = maybe [] (\x -> [(packString "receipt.id",FragmentText x)]) $ getFragmentString p (f_messageId)
	    (pf,_) <- getPresenceFrags
	    return $ np { cats = [catParseNew (unpackPS r)], fragments = fragments np ++ pf ++ rid ++ [(packString "answer.receipt",FragmentText $ packString gid)]}
	presenceLoop = do
	    let loopIsIdle ct = do
                    ct' <- waitJVarEq userActionTime ct
                    notIdle <- sendPresence True
                    if notIdle then loopNotIdle else loopIsIdle ct'
		loopNotIdle = do
		    notIdle <- sendPresence False
		    threadDelay $ (fromIntegral idleThreshold) * 1000000 
		    if notIdle then loopNotIdle else readVal userActionTime >>= loopIsIdle
	    sendPresence True
	    loopNotIdle 
	sendPresence sendIfNotIdle = do
	    (n,d) <- fmap catParseNew getGaleId
	    (pf,notIdle) <- getPresenceFrags
	    when (sendIfNotIdle || not notIdle) $ do
		p <- puffTemplate
		galeSendPuff gc $ p {cats = [( ("_gale.notice." ++ n),d)], fragments = fragments p ++ pf}
	    return notIdle
	gonePresencePuff = do
	    (n,d) <- fmap catParseNew getGaleId
	    p <- puffTemplate
	    let ef = [(f_noticePresence' ,FragmentText f_outGone), (f_noticePresence, FragmentText f_outGone)]
	    return $ p {cats = [("_gale.notice." ++ n,d)], fragments = fragments p ++ ef }
	modifyFilter f = do
	    mapVal filter_r f
	    select_perhaps select_next >> select_perhaps select_prev
	    center_puff
	center_puff = do
	    (ys,_) <- scrSize
	    selected <- readVal selected_r
	    ps <- liftM reverse getFiltered
	    let f t (((n,_),(h,_)):_) | n == selected = (t,t+h)
		f t ((_,(h,_)):rest) = f (t + h) rest
		f _ [] = (1000000,0)
	    let (mn,mx) = (f 0) ps
	    mapVal yor (min mn)
	    mapVal yor (max (mx - (ys - 1) ))
	is_at_end = do
	    (ys,_) <- scrSize
	    bs <- buf_size
	    fmap ((max (bs - (ys - 1)) 0) ==) $ readVal yor 
        scroll_end = do
	    (ys,_) <- scrSize
	    bs <- buf_size
	    writeVal yor (max (bs - (ys - 1)) 0)
	doRedraw _ (xs,ys) = do 
	    bs <- buf_size
	    mapVal yor (min (max 0 (bs - (ys - 1))))
	    selected <- readVal selected_r
	    ps <- getFiltered
	    yo <- readVal yor
	    let rp _ [] =  return () 
		rp y (((i,_),(n,f)):rest) = do
		    when ((0,ys) `overlaps` (y - yo, y - yo + n)) $ f stdScr (y - yo) (xs,ys) (i == selected)
		    rp (y + n) rest
	    attempt (rp 0 (reverse ps))
	select_perhaps a = do
	    ps <- getFiltered
	    n <- readVal selected_r
	    unless (maybe False (const True) $ lookup n (fsts ps)) a
	select_prev = do
	    ps <- liftM fsts getFiltered
	    let f sel ((n,_):_) | n < sel = n 
		f sel (_:ps) = f sel ps
		f sel [] = sel
	    mapVal selected_r (\s ->(f s ps))
	select_next = do
	    ps <- liftM fsts getFiltered
	    let f sel _ ((n,_):ps) | n > sel = f sel n ps
		f sel bg ((n,_):ps) | n > sel = f sel bg ps
		f _ bg _ = bg
	    mapVal selected_r (\s ->(f s s ps))

    let fw = widgetVerticalBox False [(ExpandFill, mwidget), (NoExpand, widgetAttr [AttrBold] sb), (NoExpand, fsw)]
	mwidget = widgetEmpty { render = rnd,  processKey = pk}
	rnd canvas = doRedraw (origin canvas) (bounds canvas)
	continue = return True
        keyTable' = (map (\(a,b) -> (a, perform_action b)) keyTable)
        pk x = case lookup x keyTable' of 
            Nothing -> case x of 
                (KeyChar n) | n `elem` "123456789" -> gfs (ord n - ord '1') >> continue
                key -> keyError "Invalid key" key >> continue
            Just x -> x
        perform_action x = case x of
                "reconnect_to_servers" -> reloadConfigFiles >> getGaleProxy >>= \gp -> galeSetProxys gc gp >>  reconnectGaleContext gc >> continue
                "ask_quit" -> 
                    let pk x = case x of 
                            (KeyChar 'y') -> return False
                            (KeyChar 'Y') -> return False
                            _ -> setRenderWidget rc fw >> return True 
                        in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk (dialog "really quit y/n?"), fw]) >> return True
                "fast_quit" -> return False
                "next_line" -> scrollPuffs 1 >> continue
                "previous_line" -> scrollPuffs (-1) >> continue
                "forward_half_page" -> scrSize >>= \(ys,_) ->  scrollPuffs (ys `div` 2) >> continue
                "backward_half_page" -> scrSize >>= \(ys,_) ->  scrollPuffs (- (ys `div` 2)) >> continue
                "next_page" -> scrSize >>= \(ys,_) -> scrollPuffs (ys) >> continue
                "previous_page" -> scrSize >>= \(ys,_) -> scrollPuffs (-ys) >> continue
                "edit_config_file" -> do
                    gc <- galeFile "ginsu.config"
                    e <- getEditor
                    mySystem (e ++ " " ++ gc)
                    reloadConfigFiles
                    touchRenderContext rc
                    return True
                "first_puff" -> writeVal selected_r 0 >> select_perhaps select_next >> writeVal yor 0 >> continue 
                "last_puff" -> do
                    ps <- readVal psr
                    case ps of 
                        ((n,_):_) -> writeVal selected_r n >> select_perhaps select_prev >> scroll_end >> continue 
                        _ -> continue
                "next_puff" -> select_next >> center_puff >> continue
                "previous_puff" -> select_prev >> center_puff >> continue
                "pop_one_filter" -> modifyFilter  (drop 1) >> continue
                "pop_all_filters" -> modifyFilter (const []) >> continue
                "invert_filter" -> modifyFilter (apHead f) >> continue where
                    f (FilterNot x) = x
                    f x = FilterNot x
                "swap_filters" -> modifyFilter f >> continue where
                    f (x:y:r) = y:x:r
                    f x = x
                "filter_current_thread" -> filter_thread >> continue
                "recall_combine_mark" -> 
                    let pk x = case x of
                                (KeyChar n) | n `elem` "123456789" -> do
                                    mark_set <- readVal mark_set_r
                                    modifyFilter ((mark_set ! (ord n - ord '1')) ++)
                                key -> keyError "Invalid mark" key  
                            >> do
                            setRenderWidget rc fw
                            return True  
                       in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk  (dialog "recall which mark? [1-9]"), fw]) >> return True 
                "set_mark" -> 
                    let pk x = case x of
                                (KeyChar n) | n `elem` "123456789" -> do
                                    fs <- readVal filter_r
                                    mapVal mark_set_r (// ([(ord n - ord '1',fs)])) 
                                    setMessage $ "mark " ++ [n] ++ " set."
                                key -> keyError "Invalid mark" key  
                            >> do
                            setRenderWidget rc fw
                            return True  
                       in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk  (dialog "set which mark? [1-9]"), fw]) >> return True 
                "filter_current_author" -> do
                    p <- readVal selPuff
                    case p of 
                        Nothing -> continue
                        Just p -> case getSigner p of
                            Just a -> do
                                modifyFilter (FilterAuthor a:)
                                continue
                            Nothing -> setMessage "no signed puff to get author from" >> continue
                "prompt_new_filter" -> do
                    (ys,_) <- scrSize
                    v <- commandRead justGetKey stdScr (ys - 1) "Filter: " ""
                    case v >>= parseFilter >>= \x -> return (modifyFilter (x:)) of
                        Just a -> a
                        Nothing -> setMessage "Invalid filter"
                    continue
                "prompt_new_filter_twiddle" -> do
                    (ys,_) <- scrSize
                    v <- commandRead justGetKey stdScr (ys - 1) "Filter: " "~"
                    case v >>= parseFilter >>= \x -> return (modifyFilter (x:)) of
                        Just a -> a
                        Nothing -> setMessage "Invalid filter"
                    continue
                "show_puff_details" -> do
                    p <- readVal selPuff
                    case p of 
                        Nothing -> setMessage "No puff selected" >> continue
                        Just p -> do
                            pdw <- puffDetailsWidget p
                            setRenderWidget rc (keyCatcherWidget (\_ -> setRenderWidget rc fw >> return True) pdw) >> continue
                "new_puff" -> composePuff (setRenderWidget rc fw)[] [] >> continue
                "modify_presence_string" -> do
                    p <- readVal myPresence
                    (ys,_) <- scrSize
                    v <- commandRead justGetKey stdScr (ys - 1) "Presence: " p
                    case v of 
                        Just v -> writeVal myPresence v >> setMessage ("Presence Updated: " ++ v)
                        Nothing -> setMessage "Presence unchanged."
                    continue
                "reply_to_author" -> do
                    p <- readVal selPuff
                    case p of
                        Nothing -> return ()
                        Just p -> do
                            pkw <- fmap (concat . (map words)) $ configLookupList "PRESERVED_KEYWORDS" 
                            let kws = filter (`elem` (map packString pkw)) (getFragmentStrings p f_messageKeyword)
                            composePuff(setRenderWidget rc fw) [catParseNew (getAuthor p)] kws 
                    continue
                "follow_up" -> do
                    p <- readVal selPuff
                    case p of
                        Nothing -> setMessage "No puff selected" 
                        Just p -> do
                            pkw <- fmap (concat . (map words)) $ configLookupList "PRESERVED_KEYWORDS" 
                            let kws = filter (`elem` (map packString pkw)) (getFragmentStrings p f_messageKeyword)
                            composePuff (setRenderWidget rc fw) (cats p) kws
                    continue
                "group_reply" -> do
                    p <- readVal selPuff
                    case p of
                        Nothing -> setMessage "No puff selected" 
                        Just p -> do
                            pkw <- fmap (concat . (map words)) $ configLookupList "PRESERVED_KEYWORDS" 
                            let kws = filter (`elem` (map packString pkw)) (getFragmentStrings p f_messageKeyword)
                            composePuff (setRenderWidget rc fw) (nub $ catParseNew (getAuthor p):cats p) kws
                    continue
                "resend_puff" -> do
                    p <- readVal selPuff
                    case p of 
                        Nothing -> setMessage "No puff selected"
                        Just p -> do
                            pcw <- puffConfirm undefined gc (setRenderWidget rc fw) p
                            setRenderWidget rc pcw
                    continue
		"redraw_screen" -> attempt (touchWin stdScr)  >> resizeRenderContext rc (return ()) >> continue
		"show_help_screen" -> setRenderWidget rc (keyCatcherWidget (\_ -> setRenderWidget rc fw >> return True) hw) >> continue
		"show_presence_status" -> setRenderWidget rc (keyCatcherWidget (\_ -> setRenderWidget rc fw >> return True) presenceWidget) >> continue
                act -> setMessage ("Unknown action: " ++ act ++ " -- " ++ helpText) >> continue
	puffDetailsWidget p = do
	    (_,xs) <- scrSize
	    let pv = newSVarWidget presence_r (presenceViewUser (getAuthor p))
            wr <- case fmap unpackPS (getFragmentString p (f_messageId)) of
                    Nothing -> return ""
                    Just mid ->  Hash.lookup puffRRHash mid >>= \ml -> case ml of 
                        Nothing -> return ""
                        Just ml -> return $ "\nReceieved By:\n" ++ unlines (snub ml)
	    v <- widgetScroll (widgetVerticalBox False [(NoExpand,widgetText $ chunkText (xs - 1) ((showPuff p) ++ wr)), (NoExpand, widgetText "\n\n\n"), (NoExpand, pv)])
	    return v
	justGetKey = fmapLeft keyCanon (readChan ic) >>= \v -> case v of
	    (Right p) -> addPuff p >> justGetKey
	    (Left k) -> do
		    getClockTime >>= writeVal userActionTime
		    return k
	nextKey = do
	    ce <- isEmptyChan ic
	    when ce $ tryDrawRenderContext rc
	    v <- fmapLeft keyCanon $ readChan ic
	    case v of 
		(Right p) -> do
                    v <- is_at_end
		    addPuff p
                    when v scroll_end
		    s <- configLookupElse "ON_INCOMING_PUFF" ""
		    mapM_ (processKey mwidget) (stringToKeys s)
		    nextKey
		(Left KeyResize) -> {-signal resizeSignal () >>-} nextKey
		(Left k) -> do
		    getClockTime >>= writeVal userActionTime
		    b <- keyRenderContext rc k 
		    if b then nextKey else return ()
    np <- configLookupBool "NO_PRESENCE_NOTIFY"

    done <- if np then return (return ()) else do 
        gonePresencePuff >>= galeWillPuff gc 
        ploopID <- forkIO presenceLoop
        return $ killThread ploopID
    forkIO $ let f n = waitJVar myPresence  n >>= \n' -> sendPresence True >> f n' in f presence
    setRenderWidget rc fw
    nextKey
    done
    

-- makeURLRegex :: IO Regex 
-- makeURLRegex = regcomp "^(([^:/?#]+):)?(//([^/?#]*))?([^?#]*)(\\?([^#]*))?(#(.*))?" regExtended


rPuff :: Int -> Puff -> (Int, Window -> Int -> (Int,Int) -> Bool -> IO ())
rPuff mw p@(Puff {cats =cats, fragments = frags}) = (n + 2,f) where
    reformat s | length s < mw = [s]
    reformat s = let (a,b) = splitAt mw s in a : reformat b
    sender = maybe "" unpackPS $ getFragmentString p f_messageSender
    body' = maybe [] (lines . paragraphBreak mw . expandTabs ) $ fmap unpackPS (getFragmentString frags  f_messageBody)
    time = getFragmentTime frags f_idTime `mplus` getFragmentTime frags (packString "_ginsu.timestamp")
    kwds = getFragmentStrings p f_messageKeyword
    --body' = map (paragraphBreak mw . expandTabs) body
    n = length body'
    addCat w (x,y) = wAttrOn w attrBold >> waddstr w x >> wAttrOff w attrBold >> waddstr w ('@':y) 
    f w y (_,ys) selected = do
	when (y >= 0 && y < ys) $ do
	    wmove w y 0
	    --wAttrOn w attrBold
	    --waddstr w (fmtS "%-20.20s" sender)
	    --wAttrOff w attrBold
	    when selected $ standout >> return ()
	    sequence (intersperse (space w) $ map (addCat w) cats)
	    space w >> space w
            waddstr w (unwords (map (('/':) . unpackPS)  kwds))
            st <- case time of 
                Just t -> liftM ft $ toCalendarTime t
                Nothing -> return ""
            let lst = length st
            wAttrOn w attrBold
            mvwaddstr w y (mw - length sender - lst - 1) sender
            wAttrOff w attrBold
            --when selected $ void standout
            mvwaddstr w y (mw - lst) st
            when selected $  standend >> return ()
        mapM_ (\(p,s) ->  when (y + p < ys) (mvwaddstr stdScr (y + p) 0 s)) (zip [1..] $ body')
        when selected $ wAttrOn w attrBold
        let bs = (if selected then "======<" else "------<") ++ getAuthor p ++ ">" ++ cycle (if selected then "=" else "-")
        --when (y + 1 + n < ys) $ mvwaddstr w (y + 1 + n) 0 (take mw $ cycle $ if selected then "=" else  "-")
        when (y + 1 + n < ys) $ mvwaddstr w (y + 1 + n) 0 (take mw bs)
        when selected $ wAttrOff w attrBold
        return ()
    ft (CalendarTime {ctHour = h, ctMin = m, ctSec = s}) = fn h ++ ":" ++ fn m ++ ":" ++ fn s
    fn n = let x = show n ; lx = length x in if lx < 2 then replicate (2 - lx) '0' ++ x else x



------------------
-- Curses routines
------------------

waddstr w s = Control.Exception.try (wAddStr w s) >> return ()
mvwaddstr w y x s = Control.Exception.try (mvWAddStr w y x s) >> return ()
wmove w y x = Control.Exception.try (wMove w y x) >> return ()

space w = waddstr w " "

--------------
-- help screen
--------------


helpWidget = do
    let w = widgetText helpTable
        s = "Help Screen"
        v = "* scroll with direction keys - any other key to return to main screen *"
    sw <- widgetScroll w 
    return $ widgetVerticalBox False [(NoExpand,widgetAttr [AttrBold] (widgetText s)),(NoExpand,widgetText v),(ExpandFill, widgetSimpleFrame sw)]

-------------------
-- Puff composition
-------------------

withPrivateFiles action = do
    om <- Posix.setFileCreationMask (Posix.groupModes `Posix.unionFileModes` Posix.otherModes)
    v <- action 
    Posix.setFileCreationMask om
    return v

noBodyWords fl = [x|x@(n,_) <- fl, n /= f_messageBody, n /= f_messageKeyword]

mySystem s = do
    putLog LogNotice $ "system " ++ show s
    withProgram $ System.system s


editPuff :: Puff -> IO (Maybe Puff)
editPuff puff = do
    e <- getEditor
    fn <- getTmpFile
    eob <- configLookupList "EDITOR_OPTION"
    eonew <- configLookupList "EDITOR_NEWPUFF_OPTION"
    let eo = if null (cats puff) then eonew else eob
    let it = ["To: " ++ simpleQuote (showDestination (cats puff) (map unpackPS $ getFragmentStrings puff f_messageKeyword)) , "--------"] 
    let mb = case fmap unpackPS (getFragmentString puff f_messageBody) of
            Just mb -> mb
            Nothing -> "\n"
    withPrivateFiles $ writeRawFile fn (stringToBytes $ unlines it ++ mb)
    mySystem (e ++ " " ++ unwords eo ++ " " ++ shellQuote [fn])
    pn <- fmap (lines . bytesToString)$ readRawFile fn 
    handleMost (\_ -> return ()) (removeFile fn) 
    if not (length pn > 1 && pn /= it) then return Nothing else do 
        let (cs',kwds') = readDestination (drop 4 (head pn))
        ncats <- expandAliases (cs')
        tw <- configLookupBool "TRIM_BLANKLINES"
        body <- return $ if not tw then (unlines $ (drop 2 pn)) else
            trimBlankLines (unlines $ (drop 2 pn))
        return $ if body == "" then Nothing else 
                Just puff { cats = ncats, fragments = noBodyWords (fragments puff) ++ [(f_messageBody,FragmentText (packString body))] ++ [ (f_messageKeyword,FragmentText (packString k)) | k <- kwds']}

showDestination cs kwds = (map catShowNew cs ++ map ('/':) kwds)
readDestination :: String -> ([Category],[String])
readDestination s = splitEither (map pi (simpleUnquote s)) where
    pi ('/':s) = Right s
    pi x = Left (catParseNew x)

--------------------
-- Puff confirmation
--------------------

anonymousFragments = map (liftT2 (packString,\x -> FragmentText (packString x))) [("id/class", "anonymous"), ("id/instance",  "anonymous"), ("message/sender",  "Anonymous")]

minusFrag fl s = [x |x@(n,_) <- fl, n /= s]

prettyPuff puff = unlines xs ++  body where
    to = ["To: " ++ simpleQuote (showDestination (cats puff) (map unpackPS $ getFragmentStrings puff f_messageKeyword)) ] 
    from = ["From: " ++ unpackPS t| (n,FragmentText t) <- fragments puff,  n == f_messageSender]
    body = case [ unpackPS t | (n,FragmentText t) <- fragments puff, n == f_messageBody] of
        (t:_) -> t
        [] -> ""
    rr = if hasFragment puff (packString "question.receipt") then ["Return Receipt: Yes"] else []
    xs = to ++ from ++ rr ++ ["-------"]


puffConfirm :: (Puff -> IO ()) -> GaleContext -> IO () -> Puff -> IO Widget
puffConfirm _ gc done puff = do
    (_,xs) <- scrSize
    --ptf <- configLookup "PUFF_TEXT_FILTER"
    gid <- getGaleId
    ncats <- expandAliases (cats puff)
    puff <- return $ puff {cats = ncats}
    psv <- newMVar puff
    let (n,d) = catParseNew gid
    let rr_cat = case getFragmentString puff (f_messageId) of 
            Just mid -> "_gale.rr." ++ n ++ "." ++ unpackPS mid ++ "@" ++ d
            Nothing -> gid

    gs <- fmap (concat . map words) $ configLookupList "GALE_SUBSCRIBE"
    when  (not (or [c `subCategory` catParseNew g | c <- cats puff, g <- gs] ) || packString "ping" `elem` getFragmentStrings puff f_messageKeyword) $ 
            writeVal psv (puff {fragments = (packString "question.receipt", FragmentText (packString rr_cat)) : fragments puff})
    sbsv <- newMVar ""
    psvs <- newMVar (prettyPuff, showPuff)
    csv <- combineVal psv psvs
    let pw = newSVarWidget csv (\(p,(f,_)) -> widgetText (chunkText (xs - 4) (f p)))
    let w = (widgetCenter $  widgetText "SEND PUFF?")
        h = (widgetCenter $ widgetText (paragraph xs $ "y:send q:abort e:edit r:returnRecipet h:allHeaders A:anonymize" ++ if False then " F:toggle filter" else ""))
        pk (KeyChar 'q')  = done >> return True
        pk (KeyChar 'n')  = done >> return True
        pk (KeyChar 'r') = do
            p <- readVal psv
            if hasFragment p (packString "question.receipt") then 
                writeVal psv (p {fragments = [x | x@(n,_) <- fragments p , n /= packString "question.receipt"]})
                    else writeVal psv (p {fragments = (packString "question.receipt", FragmentText (packString rr_cat)) : fragments p})
            return True
        pk (KeyChar 'h') = do
            mapVal psvs (\(x,y) -> (y,x))
            return True
        pk (KeyChar 'A') = do
            p <- readVal psv
            if hasFragment p (f_messageId) then do
                writeVal psv (p {signature = [], fragments = anonymousFragments `mergeFrags` (fragments p `minusFrag` f_messageId)})
                    else do
                        pt <- puffTemplate
                        writeVal psv (p {signature = signature pt, fragments = (fragments pt `minusFrag` f_idTime) `mergeFrags` fragments p})
            return True
        pk (KeyChar 'y')  = do
            puff <- readVal psv
            galeSendPuff gc puff
            done
            return True
        pk (KeyChar 'e') = do
            puff <- readVal psv
            np <- editPuff puff
            case np of
                Just np -> writeVal psv np
                Nothing -> return ()
            return True
        pk key = do
            writeVal sbsv $  "Invalid keystroke: " ++ keysToString [key]
            return True
    pb <- widgetScroll $ pw
    return $ keyCatcherWidget pk $ widgetVerticalBox False [(NoExpand, w), (NoExpand, widgetAttr [AttrBold] h), (ExpandFill, widgetSimpleFrame pb),(NoExpand, simpleStatusBarWidget sbsv)]

simpleStatusBarWidget svm  = widgetAttr [AttrBold] $ newSVarWidget svm widgetText




commandRead :: IO Curses.Key -> Window -> Int -> String -> String -> IO (Maybe String)
commandRead ic win yloc prompt init = withCursor CursorVisible  $ cr (length init) (reverse init) 
    where
    l n v = min (max n 0) (length v)
    cr cloc v = pc cloc v >> ic >>= \x -> case x of
	(KeyChar '\n') -> return $ Just (reverse v)
	(KeyChar '\r') -> return $ Just (reverse v)
	(KeyEnter) -> return $ Just (reverse v)
	(KeyChar '\b') -> let z = (let (a,b) = splitAt (length v - cloc) v in a ++ drop 1 b) in  cr (l (cloc - 1) z) z
	(KeyBackspace) -> let z = (let (a,b) = splitAt (length v - cloc) v in a ++ drop 1 b) in  cr (l (cloc - 1) z) z
        (KeyHome) -> cr 0 v
        (KeyEnd) -> cr (length v) v
        (KeyLeft) -> cr (l (cloc - 1) v) v
        (KeyRight) -> cr (l (cloc + 1) v) v
        (KeyChar '\x0B') -> cr cloc (drop (length v - cloc) v)
	(KeyChar c) -> cr (cloc + 1) (let (a,b) = splitAt (length v - cloc) v in a ++ [c] ++ b)

	_ -> return Nothing
    pc cloc v = do
	wmove win yloc 0
	wClrToEol win
	wAttrOn win attrBold
	waddstr win prompt 
	wAttrOff win attrBold
	waddstr win (reverse v)
	wmove win yloc (cloc + length prompt)
	refresh

{-
	pk x = case x of 
	    (KeyChar '\x12') -> reloadConfigFiles >> getGaleProxy >>= \gp -> galeSetProxys gc gp >>  reconnectGaleContext gc >> continue
	    (KeyChar 'q') -> 
		let pk x = case x of 
			(KeyChar 'y') -> return False
			(KeyChar 'Y') -> return False
			_ -> setRenderWidget rc fw >> return True 
		    in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk (dialog "really quit y/n?"), fw]) >> return True
	    (KeyChar 'Q') -> return False
	    (KeyChar '\n') -> scrollPuffs 1 >> continue
	    (KeyChar '\b') -> scrollPuffs (-1) >> continue
	    (KeyChar '\x05') -> scrollPuffs 1  >> continue
	    (KeyChar '\x19') -> scrollPuffs (-1)  >> continue
	    (KeyChar '\EOT') -> scrSize >>= \(ys,_) ->  scrollPuffs (ys `div` 2) >> continue
	    (KeyChar '\NAK') -> scrSize >>= \(ys,_) ->  scrollPuffs (- (ys `div` 2)) >> continue
	    (KeyChar '\ACK') -> scrSize >>= \(ys,_) ->  scrollPuffs (ys) >> continue
	    (KeyChar '\STX') -> scrSize >>= \(ys,_) ->  scrollPuffs (-ys) >> continue
	    KeyNPage -> scrSize >>= \(ys,_) -> scrollPuffs (ys) >> continue
	    KeyPPage -> scrSize >>= \(ys,_) -> scrollPuffs (-ys) >> continue
	    (KeyChar 'E') -> do
		gc <- galeFile "ginsu.config"
		e <- getEditor
		mySystem (e ++ " " ++ gc)
                reloadConfigFiles
		touchRenderContext rc
		return True
	    (KeyHome) -> writeVal selected_r 0 >> select_perhaps select_next >> writeVal yor 0 >> rcontinue 
	    (KeyEnd) -> do
		ps <- readVal psr
		case ps of 
		    ((n,_):_) -> writeVal selected_r n >> select_perhaps select_prev >> scroll_end >> rcontinue 
		    _ -> continue
	    (KeyChar 'j') -> select_next >> center_puff >> rcontinue
	    (KeyChar 'k') -> select_prev >> center_puff >> rcontinue
	    KeyDown -> select_next >> center_puff >> rcontinue
	    KeyUp -> select_prev >> center_puff >> rcontinue
	    (KeyChar 'u') -> modifyFilter  (drop 1) >> rcontinue
	    (KeyChar 'U') -> modifyFilter (const []) >> rcontinue
	    (KeyChar '!') -> modifyFilter (apHead f) >> rcontinue where
		f (FilterNot x) = x
		f x = FilterNot x
	    (KeyChar 'x') -> modifyFilter f >> rcontinue where
		f (x:y:r) = y:x:r
		f x = x
	    (KeyChar 't') -> filter_thread >> rcontinue
	    (KeyChar 'T') -> filter_thread >> rcontinue
	    (KeyChar 'C') -> 
		let pk x = case x of
			    (KeyChar n) | n `elem` "123456789" -> do
                                mark_set <- readVal mark_set_r
                                modifyFilter ((mark_set ! (ord n - ord '1')) ++)
			    key -> keyError "Invalid mark" key  
			>> do
			setRenderWidget rc fw
			return True  
		   in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk  (dialog "recall which mark? [1-9]"), fw]) >> return True 
	    (KeyChar 'm') -> 
		let pk x = case x of
			    (KeyChar n) | n `elem` "123456789" -> do
				fs <- readVal filter_r
				mapVal mark_set_r (// ([(ord n - ord '1',fs)])) 
				setMessage $ "mark " ++ [n] ++ " set."
			    key -> keyError "Invalid mark" key  
			>> do
			setRenderWidget rc fw
			return True  
		   in setRenderWidget rc (stackedWidgets [keyCatcherWidget pk  (dialog "set which mark? [1-9]"), fw]) >> return True 
	    (KeyChar 'a') -> do
		p <- readVal selPuff
		case p of 
		    Nothing -> continue
		    Just p -> case getSigner p of
			Just a -> do
			    modifyFilter (FilterAuthor a:)
			    rcontinue
			Nothing -> setMessage "no signed puff to get author from" >> continue
	    (KeyChar '/') -> do
		(ys,_) <- scrSize
		v <- commandRead justGetKey stdScr (ys - 1) "Filter: " ""
		case v >>= parseFilter >>= \x -> return (modifyFilter (x:)) of
		    Just a -> a
		    Nothing -> setMessage "Invalid filter"
		rcontinue
	    (KeyChar 'c') -> do
		(ys,_) <- scrSize
		v <- commandRead justGetKey stdScr (ys - 1) "Filter: " ""
		case v >>= parseFilter >>= \x -> return (modifyFilter (x:)) of
		    Just a -> a
		    Nothing -> setMessage "Invalid filter"
		rcontinue
	    (KeyChar '~') -> do
		(ys,_) <- scrSize
		v <- commandRead justGetKey stdScr (ys - 1) "Filter: " "~"
		case v >>= parseFilter >>= \x -> return (modifyFilter (x:)) of
		    Just a -> a
		    Nothing -> setMessage "Invalid filter"
		rcontinue
	    (KeyChar 'd') -> do
		p <- readVal selPuff
		case p of 
		    Nothing -> setMessage "No puff selected" >> continue
		    Just p -> do
			pdw <- puffDetailsWidget p
			setRenderWidget rc (keyCatcherWidget (\_ -> setRenderWidget rc fw >> return True) pdw) >> continue
	    (KeyChar 'p') -> composePuff (setRenderWidget rc fw)[] [] >> rcontinue
	    (KeyChar 'N') -> do
		p <- readVal myPresence
		(ys,_) <- scrSize
		v <- commandRead justGetKey stdScr (ys - 1) "Presence: " p
		case v of 
		    Just v -> writeVal myPresence v >> setMessage ("Presence Updated: " ++ v)
		    Nothing -> setMessage "Presence unchanged."
		continue
	    (KeyChar 'r') -> do
		p <- readVal selPuff
		case p of
		    Nothing -> return ()
		    Just p -> do
			pkw <- fmap (concat . (map words)) $ configLookupList "PRESERVED_KEYWORDS" 
			let kws = filter (`elem` (map packString pkw)) (getFragmentStrings p f_messageKeyword)
			composePuff(setRenderWidget rc fw) [catParseNew (getAuthor p)] kws 
		continue
	    (KeyChar 'f') -> do
		p <- readVal selPuff
		case p of
		    Nothing -> setMessage "No puff selected" 
		    Just p -> do
			pkw <- fmap (concat . (map words)) $ configLookupList "PRESERVED_KEYWORDS" 
			let kws = filter (`elem` (map packString pkw)) (getFragmentStrings p f_messageKeyword)
			composePuff (setRenderWidget rc fw) (cats p) kws
		continue
	    (KeyChar 'g') -> do
		p <- readVal selPuff
		case p of
		    Nothing -> setMessage "No puff selected" 
		    Just p -> do
			pkw <- fmap (concat . (map words)) $ configLookupList "PRESERVED_KEYWORDS" 
			let kws = filter (`elem` (map packString pkw)) (getFragmentStrings p f_messageKeyword)
			composePuff (setRenderWidget rc fw) (nub $ catParseNew (getAuthor p):cats p) kws
		continue
            (KeyChar 'R') -> do
                p <- readVal selPuff
                case p of 
                    Nothing -> setMessage "No puff selected"
                    Just p -> do
                        pcw <- puffConfirm undefined gc (setRenderWidget rc fw) p
                        setRenderWidget rc pcw
                continue
	    (KeyChar n) | n `elem` "123456789" -> gfs (ord n - ord '1') >> rcontinue
	    key -> keyError "Invalid key" key >> rcontinue
        -}


