module Gale(
    GaleContext,
    galeNextPuff, 
    reconnectGaleContext, 
    connectionStatus,
    dumpKey, 
    galeSendPuff, 
    hostStrings,
    galeWillPuff, 
    withGale,
    galeSetProxys,
    galeAddCategories,
    getGaleDir) where


import Char(chr,ord)
import IO hiding(bracket, bracket_)
import List
import Monad(when)
import System(getEnv)
import System.Time
import Time
import Maybe

import Network.BSD
import Network.Socket
import Data.Word(Word8, Word32)
import Control.Exception
import Control.Exception as E
import Control.Concurrent
import Data.Bits
import PackedString

import ErrorLog
import GenUtil
import Puff
import RSA
import SimpleParser
import EIO
import SHA1
import qualified Posix
import Data.Array.IO

-- TODO - prove concurrent-correctness, make sure all network errors are accounted for.

-------------------
-- Gale Constants
-------------------

galePort :: PortNumber
galePort = 11512
hostStrings s = [s, "gale." ++ s, s ++ ".gale.org."]

galeRSAModulusBits = 1024
galeRSAModulusLen = (galeRSAModulusBits + 7) `div` 8
galeRSAPrimeBits = (galeRSAModulusBits + 1) `div` 2
galeRSAPrimeLen = (galeRSAPrimeBits + 7) `div` 8

type PuffStatus = ()

data GaleContext = GaleContext { 
    connectionStatus :: !(MVar (Either String String)),
    channel :: !(Chan Puff), 
    proxy :: !(MVar [String]), 
    gThread :: ThreadId,
    gHandle :: !(MVar Handle),
    gCategory :: !(MVar [Category]),
    keyCache :: !(MVar [(Key,EvpPkey)]) 
    }

void a = a >> return ()

getGaleDir :: IO String
getGaleDir = do
    gd <- lookupEnv "GALE_DIR"
    case gd of
	Just v -> return $ v ++ "/"
	Nothing -> do
	    h <- getEnv "HOME"
	    return (h ++ "/.gale/")



-----------------
-- Implementation
-----------------

{-# NOTINLINE withGale #-}
withGale :: [String] -> (GaleContext -> IO a) -> IO a
withGale ps io = withSocketsDo $ do
    Posix.installHandler Posix.sigPIPE Posix.Ignore Nothing
    gc <- newGaleContext ps []
    r <- io gc
    destroyGaleContext gc
    return r

newGaleContext ps cs = do
    let ncs = map catParseNew cs
    cats <- newMVar $ ncs
    c <- newChan
    ps <- return (case ps of [] -> snub (concatMap (hostStrings . snd) ncs); _ -> ps)
    status <- newMVar $ Left $ "Attempting to connect to: " ++ unwords ps
    keychan <- newChan
    pmv <- newMVar ps
    hv <- newEmptyMVar
    keycachev <- newMVar []
    let gc = GaleContext { connectionStatus = status, gThread = undefined, gHandle = hv, gCategory = cats, channel = c, proxy = pmv, keyCache = keycachev}
    thd <- forkIO (connectThread gc ps hv)
    sendGimme gc
    return gc { gThread = thd }

{-# NOTINLINE galeAddCategories #-}
galeAddCategories :: GaleContext -> [Category] -> IO ()
galeAddCategories gc cs = do
    modifyMVar_ (gCategory gc) $ \cs' -> return (snub (cs ++ cs'))
    sendGimme gc

galeSetProxys :: GaleContext -> [String] -> IO ()
galeSetProxys gc ps = do
    modifyMVar_ (proxy gc) $ \cs' -> return (snub ps)
    sendGimme gc

sendGimme :: GaleContext -> IO ()
sendGimme gc = void $ forkIO $ do
    withMVar (gHandle gc) $ \h -> do
        ncs <- readMVar $ gCategory gc
        let gs = concatInter ":" (map catShowOld ncs)
        putWord32 h 2
        putWord32 h (fromIntegral $ length gs * 2)
        putRaw h $ galeEncodeString gs
        hFlush h

destroyGaleContext gc = killThread $ gThread gc

connectTo hostname port = do
    --proto <- getProtocolNumber "tcp"
    bracketOnError
	(socket AF_INET Stream 6)
	(sClose)  -- only done if there's an error
        (\sock -> do
      	  he <- getHostByName hostname
      	  connect sock (SockAddrInet port (hostAddress he))
      	  socketToHandle sock ReadWriteMode
	)

bracketOnError
	:: IO a		-- ^ computation to run first (\"acquire resource\")
	-> (a -> IO b)  -- ^ computation to run last (\"release resource\")
	-> (a -> IO c)	-- ^ computation to run in-between
	-> IO c		-- returns the value from the in-between computation
bracketOnError before after thing =
  block (do
    a <- before 
    r <- E.catch 
	   (unblock (thing a))
	   (\e -> do { after a; throw e })
    return r
 )



attemptConnect s = do
    h <- connectTo s galePort
    --hSetBuffering h NoBuffering
    return (h,s)

spc [] = []
spc s = v : spc (drop 1 r) where
    (v,r) = span (/= ':') s

emptyPuffer :: MVar Handle -> IO ()
emptyPuffer hv = repeatM_ (threadDelay 30000000 >> sendEmptyPuff) where
	sendEmptyPuff = withMVar hv $ \h -> do
		putWord32 h 0
		putWord32 h 8
		putWord32 h 0
		putWord32 h 0
		hFlush h

connectThread :: GaleContext ->  [String] -> MVar Handle -> IO ()
connectThread gc _ hv = retry 5.0 ("ConnectionError") doit where
    openHandle = do
        ds <- readMVar $ proxy gc
        swapMVar (connectionStatus gc)  $ Left $ "Attempting to connect to: " ++ unwords ds
        trySeveral (map attemptConnect ds)
    doit = bracket openHandle (hClose . fst) $ \(h,hn) -> do
	putWord32 h 1
	v <- readWord32 h
        sendGimme gc
        swapMVar (connectionStatus gc) $ Right hn
	bracket_ (putMVar hv h) (takeMVar hv) $
	  bracket (forkIO (emptyPuffer hv)) killThread $ \_ -> repeatM_ $ do
	    w <- readWord32 h
	    l <- readWord32 h
	    --d <- readRaw h (fromIntegral l)
            arr <- newArray_ (0,fromIntegral l - 1)
            dsz <- hGetArray h arr (fromIntegral l)
            when (dsz /= fromIntegral l) $ fail "short read !"
            d <- getElems arr
	    when (w == 0) $ do
		let (l,d') = xdrReadUInt d
		cat <- tryMapM parseCategoryOld (spc $ galeDecodeString (take (fromIntegral $ l ) d'))
		ct <- getClockTime
		let ef = \xs -> ((packString "_ginsu.timestamp",FragmentTime ct):(packString "_ginsu.spumbuster", FragmentText (packString (sha1ShowHash $ sha1 d))):xs)
		writeChan (channel gc) $ Puff { signature = [], cats = cat, fragments = (ef (decodeFragments $ drop (fromIntegral l + 4) d'))}

galeNextPuff :: GaleContext -> IO Puff
galeNextPuff gc = do
    p <- readChan $ channel gc
    p' <- galeDecryptPuff gc p
    putLog LogDebug $ "Puff gotten: \n" ++ (indent 4 $ showPuff p')
    return p'



reconnectGaleContext gc = do
    p <- readMVar $ proxy gc
    void $ forkIO $ attempt $ readMVar (gHandle gc) >>= hClose


galeSendPuff :: GaleContext -> Puff -> IO PuffStatus
galeSendPuff gc puff = void $ forkIO $ do
    putLog LogInfo $ "sending puff:\n" ++ (indent 4 $ showPuff puff)
    writeChan (channel gc) puff
    d <- createPuff  gc False puff
    retry 3.0 "error sending puff" $ withMVar (gHandle gc) $ \h -> putRaw h d >> hFlush h

galeWillPuff :: GaleContext -> Puff -> IO ()
galeWillPuff gc puff = void $ forkIO $ do
    putLog LogDebug $ "willing puff:\n" ++ (indent 4 $ showPuff puff)
    d <- createPuff gc True puff
    retry 3.0 "error sending puff" $ withMVar (gHandle gc) $ \h -> putRaw h d >> hFlush h


createPuff :: GaleContext -> Bool -> Puff -> IO [Word8]
createPuff gc will p@(Puff {signature = ((Signed k@(Key kn _)):sr)}) =
    getPrivateKey gc kn >>= \v -> case v of
	Nothing -> createPuff gc will $ p {signature = sr}
	Just (Key _ kfl,pkey) -> do
	    sfl <- case fragmentString f_keyOwner kfl of
		Just o -> return [(f_messageSender, FragmentText o)]
		Nothing -> return []
	    let fl = xdrWriteUInt 0 (createFragments (fragments p `mergeFrags` sfl))
	    sig <- signAll pkey fl
	    let sd = signature_magic1 ++ xdrWriteUInt (fromIntegral $ length sig) [] ++ sig ++ pubkey_magic3 ++ xdrWriteUInt (fromIntegral $ length kn) [] ++ galeEncodeString kn
	    let fd = xdrWriteUInt (fromIntegral $ length sd) $ sd ++ fl
	    createPuff gc will $ p {signature = sr, fragments = [(f_securitySignature,FragmentData fd)]}
createPuff _ will puff = do
    let cn = galeEncodeString (concatInter ":" (map catShowOld $ cats puff))
    let ad = xdrWriteUInt (fromIntegral $ length cn) (cn ++ xdrWriteUInt 0 []) ++ createFragments (fragments puff)
    let pd = xdrWriteUInt (if will then 1 else 0) (xdrWriteUInt (fromIntegral $ length ad) ad)
    evaluate pd



createFragments :: FragmentList -> [Word8]
createFragments fl = concatMap f fl where
    f (s',f) = xdrWriteUInt t $ xdrWriteUInt (fromIntegral $ length n + length xs) $ n ++ xs where
	(t, xs) = g f
	n = xdrWriteUInt (fromIntegral $ length s) (galeEncodeString s)
        s = unpackPS s'
    g (FragmentData ws) = (1, ws)
    g (FragmentText s) = (0, galeEncodeString (unpackPS s))
    g (FragmentTime (TOD s _)) = (2, replicate 4 0 ++ xdrWriteUInt (fromIntegral s) (replicate 8 0))
    g (FragmentInt i) = (3, xdrWriteUInt (fromIntegral i) [])
    g (FragmentNest fl) = (4, createFragments fl)



catfixes = [ ("/", ".|"), (".", "/"), (":", "..") ]

parseCategoryOld :: Monad m => String -> m Category
parseCategoryOld = parser p where
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(x,y) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 
    bl [] = []
    bl [x] = []
    bl (x:xs) = x:bl xs
    p = do
	char '@'
	d <- many (noneOf "/")
	parseExact "/user/"
	c <- parseRest
	return (con (bl c),d)

catShowOld :: Category -> String
catShowOld (c,d) = "@" ++ d ++ "/user/" ++ con c ++ "/" where 
    con cs | [nv] <- [x ++ (con $ drop (length y) cs) |(y,x) <- catfixes, y `isPrefixOf` cs] = nv
    con (c:cs) = c:con cs
    con "" = "" 


--------------------
-- Security routines
--------------------


unsignData :: [Word8] -> IO (FragmentList,Key)
unsignData xs = do
    key <- parseKey $ take (fromIntegral (l - (8 + sb))) (drop (fromIntegral sb) xs'')
    return (fl,key) where
	(l,xs') = xdrReadUInt xs
	(sb,xs'') = xdrReadUInt (drop 4 xs')
	fl = (decodeFragments $ drop (fromIntegral l + 4) xs')

unsignFragments :: FragmentList -> IO FragmentList
unsignFragments tfl | (xs:_) <- [xs | (n,FragmentData xs) <- tfl, n == f_securitySignature] = do
    (fl,_) <- unsignData xs
    unsignFragments fl
unsignFragments x = return x

galeDecryptPuff :: GaleContext -> Puff -> IO Puff
galeDecryptPuff gc p | (Just xs) <- getFragmentData p (f_securitySignature) = tryElse p $ do
    let (l,xs') = xdrReadUInt xs
	(sb,xs'') = xdrReadUInt (drop 4 xs')
	fl = (decodeFragments $ drop (fromIntegral l + 4) xs') ++ [f|f <- fragments p, fst f /= f_securitySignature]
    key <- parseKey $ take (fromIntegral (l - (8 + sb))) (drop (fromIntegral sb) xs'')
    galeDecryptPuff gc $ p {signature = (Unverifyable key): signature p, fragments = fl} 
galeDecryptPuff gc p | (Just xs) <- getFragmentData p f_securityEncryption = tryElse p $ do
    (cd,ks) <- parser pe xs
    dfl <- first (map (td' cd) ks) 
    let dfl' = dfl ++ [f|f <- fragments p, fst f /= f_securityEncryption]
    galeDecryptPuff gc $ p {signature = (Encrypted (map (\(_,n,_) -> n) ks)): signature p, fragments = dfl'}  where
	pe = (parseExact cipher_magic1 >> pr parseNullString) <|> (parseExact cipher_magic2 >> pr parseLenString)
	pk pkname iv = do
	    kname <- pkname
	    keydata <- parseLenData
	    return $ (iv,kname,keydata)
	pr pkname = do
	    iv <- parseSome 8
	    keycount <-  parseIntegral32
	    ks <- replicateM keycount (pk pkname iv)
	    xs <- parseRest
	    return (xs,ks)
	td' cd (iv,kname,keydata) = do
	    Just (_,pkey) <- getPrivateKey gc kname
	    dd <- decryptAll keydata iv pkey cd
	    let dfl = decodeFragments (drop 4 dd)
	    return dfl
galeDecryptPuff gc x = return x


getPrivateKey :: GaleContext -> String -> IO (Maybe (Key,EvpPkey))
getPrivateKey gc kn = modifyMVar (keyCache gc) f where
    f kc = case [x|x@(Key kn' _,_) <- kc, kn == kn'] of
	(v:_) -> return (kc,Just v)
	[] -> do
	    v <- ioMp getFromDisk
	    return ((maybeToList v ++ kc),v)
    getFromDisk = do
	gd <- getGaleDir
	let p = (gd ++ "/auth/private/")
	    knames = [p ++ kn, p ++ kn ++ ".gpri"]
	    gn = (map (\fn -> (getRawContents fn >>= \c -> return (c,fn))) knames)
	(key_c, fn) <- (first  gn) 
	key <- parseKey key_c
	rsa <- privkeyToRSA key
	pkey <- pkeyNewRSA rsa
	putLog LogNotice $ "Retrieved private key from disk: " ++ fn
	return (key,pkey)



----------------------
-- Gale stream Parsing
----------------------

parseWord32 :: GenParser Word8 Word32
parseWord32 = do
    [b1,b2,b3,b4] <- parseSome 4
    return $ (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|.
	    (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)

parseLenData = parseWord32 >>= parseSome

parseIntegral32 :: Integral a => GenParser Word8 a
parseIntegral32 = fmap fromIntegral parseWord32

parseLenString :: GenParser Word8 String
parseLenString = do
    len <- parseIntegral32
    w <- parseSome (len * 2)
    return $ galeDecodeString w

parseNullString :: GenParser Word8 String
parseNullString = (sat (== 0) >> return "") <|> (parseSome 1 >>= \[x] -> fmap ((chr $ fromIntegral x):) parseNullString)

galeDecodeString :: [Word8] -> String
galeDecodeString [] = []
galeDecodeString (b1:b2:xs) = (chr $ (fromIntegral b2) .|. (fromIntegral b1 `shiftL` 8)) : galeDecodeString xs
galeDecodeString _ = error "invalid gale string"


xdrWriteUInt :: Word32 -> [Word8] -> [Word8]
xdrWriteUInt x bs = (b1:b2:b3:b4:bs) where
    b1 = fromIntegral $ (x `shiftR` 24) .&. 0xFF
    b2 = fromIntegral $ (x `shiftR` 16) .&. 0xFF
    b3 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
    b4 = fromIntegral $ x .&. 0xFF

xdrReadUInt :: [Word8] -> ( Word32,[Word8])
xdrReadUInt (b1:b2:b3:b4:bs) = (x,bs) where
    x = (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|.
	    (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)
xdrReadUInt bs = error $ "xdrReadUInt " ++ show bs

putWord32 :: Handle -> Word32 -> IO ()
putWord32 h x = do
    hPutChar h $ chr $ fromIntegral $ (x `shiftR` 24) 
    hPutChar h $ chr $ fromIntegral $ (x `shiftR` 16) .&. 0xFF
    hPutChar h $ chr $ fromIntegral $ (x `shiftR` 8) .&. 0xFF
    hPutChar h $ chr $ fromIntegral $ x .&. 0xFF
    
    --putRaw h $ xdrWriteUInt x []

readWord32 :: Handle -> IO Word32
readWord32 h = do
    a <- newArray_ (0,3)
    n <- hGetArray h a 4 
    when (n /= 4) $ fail "short read." 
    [b1,b2,b3,b4] <- getElems a
    return $ (fromIntegral b4) .|. (fromIntegral b3 `shiftL` 8) .|.
	     (fromIntegral b2 `shiftL` 16) .|. (fromIntegral b1 `shiftL` 24)
    --xs <- readRaw h 4
    --return $ fst $ xdrReadUInt xs

galeEncodeString :: String -> [Word8]
galeEncodeString cs = concatMap (f . ord) (concat $ map (\c -> if c == '\n' then "\r\n" else [c]) cs) where
    f x = (b1:b2:[]) where
	b1 = fromIntegral $ (x `shiftR` 8) .&. 0xFF
	b2 = fromIntegral $ x .&. 0xFF

fragmentData' x y = fragmentData (packString x) y
privkeyToRSA :: Key -> IO RSA
privkeyToRSA (Key _ fl)  = do
    rsa <- rsaNew
    fragmentData' "rsa.modulus" fl >>= bn_bin2bn >>= rsaSetN rsa
    fragmentData' "rsa.exponent" fl >>= bn_bin2bn >>= rsaSetE rsa
    fragmentData' "rsa.private.exponent" fl >>= bn_bin2bn >>= rsaSetD rsa
    fragmentData' "rsa.private.coefficient" fl >>= bn_bin2bn >>= rsaSetIQMP rsa
    xs <- fragmentData' "rsa.private.prime" fl
    let (p,q) = splitAt galeRSAPrimeLen xs
    bn_bin2bn p >>= rsaSetP rsa
    bn_bin2bn q >>= rsaSetQ rsa
    xs <- fragmentData' "rsa.private.prime.exponent" fl
    let (dmp1,dmq1) = splitAt galeRSAPrimeLen xs
    bn_bin2bn dmp1 >>= rsaSetDMP1 rsa
    bn_bin2bn dmq1 >>= rsaSetDMQ1 rsa
    rsaCheckKey rsa
    return rsa


--key_response_cat = "_gale.key."
--key_request_cat = "_gale.query."

--------------
-- Key Parsing
--------------

stons :: [Char] -> [Word8]
stons = map (fromIntegral . ord)

private_magic1, private_magic2, private_magic3 :: [Word8]
cipher_magic1, cipher_magic2 :: [Word8]
pubkey_magic1, pubkey_magic2, pubkey_magic3 :: [Word8]
signature_magic1 :: [Word8]

cipher_magic1 = stons "h\DC3\002\000"
cipher_magic2 = stons "h\DC3\002\001"

signature_magic1 = stons "h\DC3\001\000"

private_magic1 = stons "h\DC3\000\001"
private_magic2 = stons "h\DC3\000\003"
private_magic3 = stons "GALE\000\002"


pubkey_magic1 = stons "h\DC3\000\000"
pubkey_magic2 = stons "h\DC3\000\002"
pubkey_magic3 = stons "GALE\000\001"


flipLocalPart :: String -> String
flipLocalPart s | '@' `notElem` s = s
flipLocalPart s = nbp ++ ep where
    nbp = concat $ reverse (groupBy f bp)
    (bp,ep) = span (/= '@') s
    f '.' '.' = True
    f x y | x /= '.' && y /= '.' = True
    f _ _ = False

getRawContents :: String -> IO [Word8]
getRawContents fn = do
    c <- readFile fn
    return (map (fromIntegral . ord) c)


keyDecode12 xs = [ (packString x,y) | (x,y) <- fl] where
    fl = [("rsa.modulus",FragmentData modulusD),
	  ("rsa.exponent",FragmentData exponentD),
	  ("rsa.private.exponent",FragmentData privateExponentD),
	  ("rsa.private.prime",FragmentData privatePrimeD),
	  ("rsa.private.prime.exponent",FragmentData privatePrimeExponentD),
	  ("rsa.private.coefficient",FragmentData privateCoefficientD),
	  ("rsa.bits", FragmentInt (fromIntegral bits))
	 ]
    (bits,ys) = xdrReadUInt xs
    (modulusD,xs') = splitRLE galeRSAModulusLen ys
    (exponentD,xs'') = splitRLE galeRSAModulusLen xs'
    (privateExponentD,xs''') = splitRLE (galeRSAPrimeLen * 2) xs''
    (privatePrimeD,xs'''') = splitRLE galeRSAModulusLen xs'''
    (privatePrimeExponentD,xs''''') = splitRLE (galeRSAPrimeLen * 2) xs''''
    (privateCoefficientD,_) = splitRLE galeRSAPrimeLen xs'''''



keyParse :: GenParser Word8 Key
keyParse = choice [ppk1,ppk2,ppk3,pk1,pk2,pk3] where
    ppk1 = do
	parseExact private_magic1
	kn <- fmap flipLocalPart parseNullString
	fl <- toParser keyDecode12
	return $ Key kn fl
    ppk2 = do
	parseExact private_magic2
	kn <- fmap flipLocalPart parseLenString
	fl <- toParser keyDecode12
	return $ Key kn fl
    ppk3 = do
	parseExact private_magic3
	kn <- fmap flipLocalPart parseLenString
	fl <- toParser decodeFragments
	return $ Key kn fl
    pk1 = do
	parseExact pubkey_magic1
	kn <- parseNullString
	(eof >> return (Key kn [])) <|> do
	comment <- parseNullString
	fl <- parsePublic12
	signature <- parseRest
	return $ Key kn ([(f_keyOwner, FragmentText (packString comment))] ++ fl)
    pk2 = do
	parseExact pubkey_magic2
	kn <- parseLenString
	(eof >> return (Key kn [])) <|> do
	comment <- parseLenString
	fl' <- parsePublic12
	let fl = ((f_keyOwner, FragmentText (packString comment)):fl')
	(eof >> return (Key kn fl)) <|> do
	ts <- parseSome 16
	te <- parseSome 16
	signature <- parseRest
	return $ Key kn $ fl ++ [(f_keySigned,FragmentTime (decodeTime ts)), (f_keyExpires,FragmentTime (decodeTime te))]
    pk3 = do
	parseExact pubkey_magic3
	kn <- parseLenString
	fl <- toParser decodeFragments
	return $ Key kn fl
    parsePublic12 = do
	bits <- parseIntegral32
	modulus <- (MkP (\x -> Just (splitRLE galeRSAModulusLen x)))
	exponent <- (MkP (\x -> Just (splitRLE galeRSAModulusLen x)))
	return [(f_rsaModulus, FragmentData modulus),
		(f_rsaExponent, FragmentData exponent),
		(f_rsaBits, FragmentInt bits)]


parseKey :: [Word8] -> IO Key
parseKey xs = do
    (Key kn fl) <- parser keyParse xs
    fl <- unsignFragments fl
    return $ Key kn fl

showKey :: Key -> String
showKey (Key n fl) = "Key: " ++ n ++ "\n" ++ (indentLines 4 $ showFragments fl)

---------------
-- data parsing
---------------

splitRLE :: Int -> [Word8] -> ([Word8], [Word8])
splitRLE n _ | n < 0 = error "invalid RLE encoding"
splitRLE 0 xs = ([],xs)
splitRLE n (c:xs) | c .&. 0x80 /= 0 = (take count xs ++ ys,rest) where
    count = fromIntegral $ (c .&. 0x7f) + 1
    (ys,rest) = splitRLE (n - count) (drop count xs)
splitRLE n (c:x:xs) | c .&. 0x80 == 0 = (replicate count x ++ ys,rest) where
    count = fromIntegral $ (c .&. 0x7f) + 1
    (ys,rest) = splitRLE (n - count) xs
splitRLE _ _  = error "invalid RLE encoding"

toParser :: ([c] -> a) -> GenParser c a
toParser f = MkP (\xs -> Just (f xs, []))

decodeFragments :: [Word8] -> FragmentList
decodeFragments [] = []
decodeFragments xs = decodeFragment ft f : decodeFragments r where
    (ft,xs') = xdrReadUInt xs
    (fl,xs'') = xdrReadUInt xs'
    (f,r) = splitAt (fromIntegral fl) xs''

decodeFragment :: Word32 -> [Word8] -> (PackedString, Fragment)
decodeFragment t xs = (fn,f t) where
    (fnl,xs') = xdrReadUInt xs
    (fn,xs'') = liftT2 (packString . galeDecodeString,id) (splitAt (fromIntegral $ fnl * 2) xs')
    f 0 = FragmentText $ packString (galeDecodeString xs'')
    f 1 = FragmentData xs''
    f 2 = FragmentTime $ decodeTime xs''
    f 3 = FragmentInt (fromIntegral $ fst $ xdrReadUInt xs'')
    f 4 = FragmentNest (decodeFragments xs'')
    f x = error $ "unknown fragment: " ++ show x

decodeTime :: [Word8] -> ClockTime
decodeTime xs = TOD (fromIntegral t) 0 where
    (t,_) = xdrReadUInt (drop 4 xs)

{-
decodeTime (x:y:_) | x > 0 || y > 0 = endOfTime
decodeTime xs = TimeDiff {tdDay = 0, tdPicosec = 0, tdYear = 0, tdMonth = 0, tdHour = 0, tdMin = fromIntegral (t `div` 60) , tdSec = fromIntegral (t `mod` 60)} `addToClockTime` epoch where
    (t,_) = xdrReadUInt (drop 4 xs)

-}


----------------
-- test code
----------------

{-# NOTINLINE dumpKey #-}
dumpKey :: String -> IO ()
dumpKey arg = do
    c <- getRawContents arg
    key <- parseKey c
    putStrLn $ showKey key


